/*-------------------------------------------------------------------------+
| Copyright 2020-2024 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.common.ui.javafx.control.treetableview;

import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableValue;
import javafx.scene.control.TextField;
import javafx.scene.control.TreeTableRow;
import javafx.scene.control.cell.TextFieldTreeTableCell;
import javafx.scene.input.KeyCode;
import javafx.util.StringConverter;
import javafx.util.converter.DefaultStringConverter;

/**
 * A DynamicTextFieldTreeTableCell that can dynamically decide per element,
 * whether it should be editable.
 * 
 * Also the editCancel behavior can be configured for cases in which the cell
 * only loses the selection.
 * 
 * @author bayha
 */
public class DynamicTextFieldTreeTableCell<T> extends TextFieldTreeTableCell<T, String> {
	/** Flag to indicate any occurred Escape key press events during editing. */
	private boolean escapePressed = false;

	/** The UIProvider of the editor in which this cell is embedded in. */
	DynamicTreeTableUIProviderBase<T> uiProvider;

	/** The column index, this cell is embedded in. */
	int columnIndex;

	/**
	 * Constructor.
	 */
	public DynamicTextFieldTreeTableCell(DynamicTreeTableUIProviderBase<T> uiProvider,
			int columnIndex) {
		this(new DefaultStringConverter());

		this.uiProvider = uiProvider;
		this.columnIndex = columnIndex;

		setConverter(new DefaultStringConverter());
	}

	/**
	 * Constructor.
	 */
	public DynamicTextFieldTreeTableCell(StringConverter<String> converter) {
		super(converter);

		this.setOnKeyPressed(event -> {
			if(event.getCode().equals(KeyCode.ESCAPE)) {
				escapePressed = true;
			} else {
				escapePressed = false;
			}
		});
	}

	/** {@inheritDoc} */
	@Override
	public void startEdit() {
		TreeTableRow<T> row = getTreeTableRow();
		T currentEntry = row.getItem();

		boolean isElementEditable = !uiProvider.isElementEditable(columnIndex, currentEntry);
		if(currentEntry == null || !isEditable() || !getTreeTableView().isEditable() ||
				!getTableColumn().isEditable() || isElementEditable) {
			return;
		}

		super.startEdit();

		TextField textField = (TextField)getGraphic();
		if(textField != null) {

			// Listener to add possibility to update the model in case the editor loses focus (user
			// clicking elsewhere or escape pressed).
			textField.focusedProperty().addListener(new ChangeListener<Boolean>() {

				@Override
				public void changed(ObservableValue<? extends Boolean> val, Boolean oldVal,
						Boolean newVal) {
					String newValue = textField.getText();
					String oldValue = uiProvider.getLabel(getTreeTableRow().getItem(), columnIndex);
					// Do not update value in case it did not change or escape was pressed (i.e.
					// intended cancel).
					if(!newValue.equals(oldValue) && !escapePressed) {
						T currentlyEditedElement = getTreeTableRow().getItem();
						if(currentlyEditedElement != null) {
							uiProvider.updateValue(currentlyEditedElement, columnIndex, newValue);
						}
					}
				}
			});
		}
	}

	/** {@inheritDoc} */
	@Override
	public void cancelEdit() {
		super.cancelEdit();

		getTreeTableView().refresh();
	}

	/** {@inheritDoc} */
	@Override
	public void updateItem(String item, boolean empty) {
		super.updateItem(item, empty);

		uiProvider.styleCell(this, columnIndex);
	}
}
