/*-------------------------------------------------------------------------+
| Copyright 2023 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.common.ui.javafx.control.treetableview;

import static javafx.embed.swt.SWTFXUtils.toFXImage;

import java.lang.reflect.Constructor;
import java.util.List;

import org.eclipse.jface.resource.ImageDescriptor;

import javafx.event.ActionEvent;
import javafx.event.EventHandler;
import javafx.scene.Node;
import javafx.scene.control.ContextMenu;
import javafx.scene.control.MenuItem;
import javafx.scene.control.SeparatorMenuItem;
import javafx.scene.image.Image;
import javafx.scene.image.ImageView;

/**
 * Base class for {@link MenuItem}s used by {@link DynamicTreeViewer}s (but in principle generally
 * applicable for JavaFX {@link ContextMenu}s).
 * <ul>
 * <li>Subclasses must provide exactly one public constructor that takes the underlying tree element
 * as single argument.</li>
 * <li>Instances must be created using factory method
 * {@link #createTreeContextMenu(List, Object)}.</li>
 * </ul>
 */
public abstract class TreeContextMenuItem<T> extends MenuItem {

	/** The model element on which the context menu operates. */
	private T element;

	/** Constructor. */
	protected TreeContextMenuItem(T element, String label) {
		super(label);
		init(element);
	}

	/** Constructor. */
	protected TreeContextMenuItem(T element, String label, Node graphic) {
		super(label, graphic);
		init(element);
	}

	/** Constructor. */
	protected TreeContextMenuItem(T element, String label, ImageDescriptor imgDescr) {
		this(element, label, getSelectIcon(imgDescr));
	}

	/** Helper to build the select icon. */
	private static Node getSelectIcon(ImageDescriptor imgDescr) {

		Image fxIcon = toFXImage(imgDescr.getImageData(100), null);
		return new ImageView(fxIcon);
	}

	/** Helper for constructors. */
	private void init(T element) {
		this.element = element;
		setOnAction(createOnActionHandler());
	}

	/** Getter for {@link #element}. */
	protected T getElement() {
		return element;
	}

	/** Predicate iff this {@link TreeContextMenuItem} is hidden. */
	public abstract boolean isHidden();

	/** Predicate when this {@link TreeContextMenuItem} is disabled. */
	public abstract boolean isDisabled();

	/**
	 * Factory method for {@link EventHandler} that implements the action triggered by the
	 * context menu.
	 */
	protected abstract EventHandler<ActionEvent> createOnActionHandler();

	/**
	 * Factory method for a {@link ContextMenu} defined based on {@link TreeContextMenuItem}s.
	 * {@code null} may be used to denote menu separators.
	 */
	public static <T> ContextMenu createTreeContextMenu(
			List<Class<? extends TreeContextMenuItem<T>>> contextMenuEntryTypes, T element) {

		boolean previousWasItem = false;
		boolean needSeparator = false;
		ContextMenu menu = new ContextMenu();

		for(Class<? extends TreeContextMenuItem<T>> entryType : contextMenuEntryTypes) {
			if(entryType == null) {
				if(previousWasItem) {
					// Separator is only needed if there is a preceding visible item
					needSeparator = true;
					previousWasItem = false;
				}
			} else {
				try {
					Constructor<?>[] ctrs = entryType.getConstructors();
					if(ctrs.length != 1) {
						throw new RuntimeException(
								"Subclasses of TreeContextMenuItem must provide exactly one public constructor");
					}
					@SuppressWarnings("unchecked") Constructor<? extends TreeContextMenuItem<T>> ctr =
							(Constructor<? extends TreeContextMenuItem<T>>)ctrs[0];
					if(ctr.getParameterCount() != 1) {
						throw new RuntimeException(
								"The constructor of subclasses of TreeContextMenuItem must take the tree element for which the context menu is created as single argument.");
					}

					TreeContextMenuItem<?> entry = ctr.newInstance(element);
					if(!entry.isHidden()) {
						// Add separator only if there is a subsequent visible item
						if(needSeparator) {
							menu.getItems().add(new SeparatorMenuItem());
							needSeparator = false;
						}
						menu.getItems().add(entry);
						entry.setDisable(entry.isDisabled());
						previousWasItem = true;
					}
				} catch(Exception e) {
					e.printStackTrace();
				}
			}
		}

		return menu;
	}
}
