/*-------------------------------------------------------------------------+
| Copyright 2023 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.reuse.ui.commands;

import static org.conqat.ide.commons.ui.ui.WorkbenchUtils.getActiveWorkbenchWindow;
import static org.eclipse.jface.dialogs.MessageDialog.openError;
import static org.eclipse.jface.window.Window.OK;
import static org.eclipse.ui.PlatformUI.getWorkbench;
import static org.fortiss.tooling.ext.reuse.ui.prototypes.PrototypeProvider.DEFAULT_ROOT_REUSE_ELEMENT_NAME;
import static org.fortiss.tooling.ext.reuse.ui.utils.ReuseLibraryUIUtils.manageExternalReferencesOfReuseElement;
import static org.fortiss.tooling.ext.reuse.ui.utils.ReuseLibraryUIUtils.warnOfDeletedReuseElem;
import static org.fortiss.tooling.ext.reuse.ui.utils.ReuseLibraryViewUtils.raiseAndUpdateReuseLibraryView;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsBasics.doesReuseElementExistWithContext;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsBasics.getAllLocalReuseLibraries;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsBasics.getReuseElementName;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsManipulation.addElementToLibrary;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsManipulation.createAndAddReuseLibrary;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsManipulation.createElementForLibrary;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.jface.action.Action;
import org.eclipse.jface.dialogs.InputDialog;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Shell;
import org.fortiss.tooling.ext.reuse.model.ReuseLibrary;
import org.fortiss.tooling.ext.reuse.ui.ToolingReuseUIActivator;
import org.fortiss.tooling.ext.reuse.ui.dialog.AddElementToLibDialog;
import org.fortiss.tooling.kernel.internal.PrototypeService;
import org.fortiss.tooling.kernel.service.ICommandStackService;

/**
 * Action to add an element to a {@link ReuseLibrary}.
 *
 * @author bergemann
 */
public class AddElementToLibraryAction extends Action {
	/** The title of the dialog shell(s). */
	private static final String SHELL_TITLE = "Add Element to Reuse Library";

	/** The description of the processed action. */
	private static final String ACTION_DESCRIPTION =
			"Adding a selected reuse element to a selected reuse library";

	/** The selected element. */
	private EObject selectedElement;

	/** The selected {@link ReuseLibrary}. */
	private ReuseLibrary selectedLibrary;

	/** All names of current {@link ReuseLibrary}s. */
	private List<String> currentExistingLibraryNames;

	/** All names of current {@link ReuseLibrary} elements. */
	private List<String> currentExistingLibraryElementNames;

	/** Constructor. Takes the selected element as input. */
	public AddElementToLibraryAction(EObject element) {
		super("Add to Reuse Library",
				ToolingReuseUIActivator.getImageDescriptor("icons/library_add.png"));
		selectedElement = element;
	}

	/** {@inheritDoc} */
	@Override
	public void run() {
		// TODO #4302: This method is quite long and should be split up.
		selectedLibrary = null;
		currentExistingLibraryNames = new ArrayList<String>();
		List<ReuseLibrary> libraryList = getAllLocalReuseLibraries();

		Boolean createNewLibrary;
		String creationText;
		if(libraryList != null && !libraryList.isEmpty()) {
			// Update existing library names.
			for(ReuseLibrary tmpLibrary : libraryList) {
				currentExistingLibraryNames.add(getReuseElementName(tmpLibrary));
			}

			// Ask for library selection.
			final Shell shell = getWorkbench().getActiveWorkbenchWindow().getShell();
			final AddElementToLibDialog librarySelectDialog =
					new AddElementToLibDialog(shell, SHELL_TITLE);
			librarySelectDialog.setLibraryList(libraryList);
			librarySelectDialog.open();

			if(!librarySelectDialog.goAhead()) {
				return;
			}
			selectedLibrary = librarySelectDialog.getLibrary();
			createNewLibrary = librarySelectDialog.createNew();
			creationText =
					"A new library will be created in which the element can be added.\nEnter a name for the library:\n ";
		} else {
			createNewLibrary = true;
			creationText =
					"No library exists yet to add the element. A new one will be created.\nEnter a name for the library:\n ";
		}

		if(createNewLibrary) {
			final InputDialog creationDialog =
					new InputDialog(Display.getCurrent().getActiveShell(), SHELL_TITLE,
							creationText, DEFAULT_ROOT_REUSE_ELEMENT_NAME,
							new NameValidator(currentExistingLibraryNames, "reuse library"));
			if(creationDialog.open() != OK) {
				return;
			}
			if(doesReuseElementExistWithContext(selectedElement)) {
				ICommandStackService.getInstance().runAsCommand(selectedElement, () -> {
					selectedLibrary = createAndAddReuseLibrary(creationDialog.getValue());
				});
			} else {
				warnOfDeletedReuseElem(ACTION_DESCRIPTION, SHELL_TITLE);
				return;
			}
		}

		if(selectedLibrary == null) {
			return;
		}

		// Update existing element names.
		currentExistingLibraryElementNames = new ArrayList<String>();
		EList<EObject> elementList = selectedLibrary.getReuseElementList();
		for(EObject libraryElement : elementList) {
			currentExistingLibraryElementNames.add(getReuseElementName(libraryElement));
		}

		// Ask for new element name (when saving element to library).
		final InputDialog addingDialog = new InputDialog(Display.getCurrent().getActiveShell(),
				SHELL_TITLE,
				"Enter the name with which the element should be saved in the library '" +
						getReuseElementName(selectedLibrary) + "':\n ",
				getReuseElementName(selectedElement),
				new NameValidator(currentExistingLibraryElementNames, "reuse library element"));
		if(addingDialog.open() != OK) {
			return;
		}
		String newName = addingDialog.getValue();

		// Execute "adding" action (if the element still exists).
		if(doesReuseElementExistWithContext(selectedElement)) {
			ICommandStackService.getInstance().runAsCommand(selectedElement, () -> {
				addElement(selectedLibrary, selectedElement, newName);
			});
		} else {
			warnOfDeletedReuseElem(ACTION_DESCRIPTION, SHELL_TITLE);
			return;
		}

		raiseAndUpdateReuseLibraryView();
	}

	/** Add one given element to specified {@link ReuseLibrary}. */
	protected void addElement(ReuseLibrary library, EObject elementToAdd, String saveName) {
		if(library == null || elementToAdd == null || saveName == null) {
			return;
		}

		EObject newElementForLibrary = createElementForLibrary(elementToAdd, library, saveName);

		try {
			addElementToLibrary(newElementForLibrary, library);
		} catch(Exception e) {
			openError(getActiveWorkbenchWindow().getShell(), "Error in '" + SHELL_TITLE + "'",
					e.getMessage());
		}

		PrototypeService.getInstance().updatePrototypes();

		String currentActionType = "addToLibrary";
		manageExternalReferencesOfReuseElement(elementToAdd, currentActionType, SHELL_TITLE);
	}
}
