/*-------------------------------------------------------------------------+
| Copyright 2023 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.reuse.ui.commands;

import static org.conqat.ide.commons.ui.dialog.MessageUtils.showInfo;
import static org.fortiss.tooling.ext.reuse.ui.commands.UpdateElementFromLibraryAction.UPDATE_FROM_CONTEXT_MENU_STRING;
import static org.fortiss.tooling.ext.reuse.ui.utils.ReuseLibraryUIUtils.warnOfDeletedReuseElem;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsBasics.doesReuseElementExistWithContext;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsBasics.getAllUpdatedOrigins;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsBasics.getFirstSourceLibraryOfElement;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsBasics.getLibraryIDName;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsBasics.getReuseElementLastUpdateString;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsBasics.getReuseElementName;

import java.util.List;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.jface.action.Action;
import org.fortiss.tooling.ext.reuse.model.ReuseLibrary;
import org.fortiss.tooling.ext.reuse.ui.ToolingReuseUIActivator;
import org.fortiss.tooling.kernel.service.ICommandStackService;

/**
 * Action to get information about updates of origin reuse elements in a {@link ReuseLibrary}.
 *
 * @author bergemann
 */
public class GetUpdatedOriginsAction extends Action {

	/** The title of the dialog shell(s). */
	private static final String SHELL_TITLE = "Information about updated reuse origins";

	/** The description of the processed action. */
	private static final String ACTION_DESCRIPTION =
			"Informing the user about updated reuse origins";

	/** The string of the context menu entry. */
	private static final String UPDATED_ORIGIN_CONTEXT_MENU_STRING = "Get reuse update information";

	/** The selected element. */
	private EObject selectedElement;

	/** Constructor. Takes the selected element as input. */
	public GetUpdatedOriginsAction(EObject element) {
		super(UPDATED_ORIGIN_CONTEXT_MENU_STRING,
				ToolingReuseUIActivator.getImageDescriptor("icons/lib_update_info.png"));
		selectedElement = element;
	}

	/** {@inheritDoc} */
	@Override
	public void run() {
		if(doesReuseElementExistWithContext(selectedElement)) {
			ICommandStackService.getInstance().runAsCommand(selectedElement, () -> {
				displayOriginUpdateInfo(selectedElement);
			});
		} else {
			warnOfDeletedReuseElem(ACTION_DESCRIPTION, SHELL_TITLE);
			return;
		}
	}

	/** Displays information about possible origin updates of the given reuse element. */
	private void displayOriginUpdateInfo(EObject reuseElement) {
		if(reuseElement == null) {
			return;
		}

		List<EObject> updatedReuseOrigins = getAllUpdatedOrigins(reuseElement);

		// Currently, it is just a brief text message about this information.
		// It could also be possible to provide directly update options within this
		// info/dialog box.
		String infoMessage;
		if(updatedReuseOrigins.isEmpty()) {
			infoMessage =
					"None of the found reuse origins was updated since the selected reuse element was reused or updated (based of these origins)." +
							"\n\nTherefore, it is not needed to get updates from any of them except you want to get the original version (e.g., if you have manipulated the selected reuse element in the meantime and want to reset it)." +
							"\n\nKeep in mind: Updated reuse origins cannot be found if your local reuse libraries are not up to date as well!";
		} else {
			infoMessage =
					"The following reuse origins of the selected reuse element were updated since the selected reuse element was reused or updated (based of these origins):\n";
			for(EObject updatedReuseOrigin : updatedReuseOrigins) {
				String reuseElementName = getReuseElementName(updatedReuseOrigin);
				ReuseLibrary library = getFirstSourceLibraryOfElement(updatedReuseOrigin);
				String libraryName = getLibraryIDName(library);
				boolean stringIsForDisplay = true;
				String lastUpdate =
						getReuseElementLastUpdateString(updatedReuseOrigin, stringIsForDisplay);
				infoMessage += "\nName: '" + reuseElementName + "'\n";
				infoMessage += "Within library: '" + libraryName + "'\n";
				infoMessage += "Last updated: " + lastUpdate + "\n";
			}
			infoMessage += "\nIf you want to get one of these updates, please use the '" +
					UPDATE_FROM_CONTEXT_MENU_STRING + "' context menu entry.";
		}
		showInfo(SHELL_TITLE, infoMessage);
	}
}
