/*-------------------------------------------------------------------------+
| Copyright 2023 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.reuse.ui.commands;

import static java.lang.String.join;
import static org.conqat.ide.commons.ui.dialog.MessageUtils.confirm;
import static org.conqat.ide.commons.ui.ui.WorkbenchUtils.getActiveWorkbenchWindow;
import static org.eclipse.jface.dialogs.MessageDialog.openError;
import static org.eclipse.jface.window.Window.OK;
import static org.fortiss.tooling.ext.reuse.storage.ReuseLibraryStorageManager.saveReuseLibrary;
import static org.fortiss.tooling.ext.reuse.ui.utils.ReuseLibraryUIUtils.checkForElementInLibrary;
import static org.fortiss.tooling.ext.reuse.ui.utils.ReuseLibraryUIUtils.manageExternalReferencesOfReuseElement;
import static org.fortiss.tooling.ext.reuse.ui.utils.ReuseLibraryUIUtils.selectCorrectReuseSpec;
import static org.fortiss.tooling.ext.reuse.ui.utils.ReuseLibraryUIUtils.warnOfDeletedReuseElem;
import static org.fortiss.tooling.ext.reuse.ui.utils.ReuseLibraryUIUtils.warnOfMissingLib;
import static org.fortiss.tooling.ext.reuse.ui.utils.ReuseLibraryViewUtils.raiseAndUpdateReuseLibraryView;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsBasics.doesReuseElementExistWithContext;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsBasics.getLibraryIDName;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsBasics.getLocalReuseLibraryByID;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsBasics.getReuseElementName;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsManipulation.replaceElementInLibrary;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getParentsWithType;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.jface.action.Action;
import org.eclipse.jface.dialogs.IInputValidator;
import org.eclipse.jface.dialogs.InputDialog;
import org.eclipse.swt.widgets.Display;
import org.fortiss.tooling.ext.reuse.model.ReuseElementSpec;
import org.fortiss.tooling.ext.reuse.model.ReuseLibrary;
import org.fortiss.tooling.ext.reuse.ui.ToolingReuseUIActivator;
import org.fortiss.tooling.kernel.internal.PrototypeService;
import org.fortiss.tooling.kernel.model.INamedCommentedElement;
import org.fortiss.tooling.kernel.service.ICommandStackService;

/**
 * Action to update a reuse element inside a {@link ReuseLibrary}.
 *
 * @author bergemann
 */
public class UpdateElementInLibraryAction extends Action {

	/** The title of the dialog shell(s). */
	private static final String SHELL_TITLE = "Update Element In Reuse Library";

	/** The description of the processed action. */
	private static final String ACTION_DESCRIPTION =
			"Updating the selected reuse element inside the selected reuse library";

	/** The string of the context menu entry. */
	private static final String UPDATE_IN_CONTEXT_MENU_STRING = "Update in Reuse Library";

	/** The selected element. */
	private EObject selectedElement;

	/** All names of current {@link ReuseLibrary} elements. */
	private List<String> currentExistingLibraryElementNames;

	/** Constructor. Takes the selected element as input. */
	public UpdateElementInLibraryAction(EObject element) {
		super(UPDATE_IN_CONTEXT_MENU_STRING,
				ToolingReuseUIActivator.getImageDescriptor("icons/update_in_library.png"));
		selectedElement = element;
	}

	/** {@inheritDoc} */
	@Override
	public void run() {
		// TODO #4302: Remove code duplication with UpdateElementFromLibraryAction.
		ReuseElementSpec correctLibrarySpec = selectCorrectReuseSpec(selectedElement, SHELL_TITLE);
		if(correctLibrarySpec == null) {
			return;
		}
		ReuseLibrary correctLibrary =
				getLocalReuseLibraryByID(correctLibrarySpec.getSourceLibUUID());

		if(correctLibrary == null) {
			warnOfMissingLib(correctLibrarySpec, SHELL_TITLE);
			return;
		}

		// Find original element inside the library.
		EObject originalLibraryElement =
				checkForElementInLibrary(correctLibrary, correctLibrarySpec, SHELL_TITLE);
		if(originalLibraryElement == null) {
			return;
		}
		String currentNameInLibrary = getReuseElementName(originalLibraryElement);
		String currentProjectName = "not found";
		EList<INamedCommentedElement> parentList =
				getParentsWithType(selectedElement, INamedCommentedElement.class);
		if(parentList != null && !parentList.isEmpty()) {
			// Safe due to check above and last element is the highest parent aka root.
			INamedCommentedElement root = parentList.get(parentList.size() - 1);
			currentProjectName = root.getName();
		}

		// Confirm the update.
		List<String> textLines = new ArrayList<String>();
		textLines.add("Confirm the following update (from source to target):");
		textLines.add("\n- Update source: element in current project");
		textLines.add("    > element name: " + getReuseElementName(selectedElement));
		textLines.add("    > in project: " + currentProjectName);
		textLines.add("\n- Update target: element in library");
		textLines.add("    > element name: " + currentNameInLibrary);
		textLines.add("    > in library: " + getLibraryIDName(correctLibrary));
		boolean confirmed = confirm(SHELL_TITLE, join("\n", textLines));

		if(!confirmed) {
			return;
		}

		// Update existing element names.
		currentExistingLibraryElementNames = new ArrayList<String>();
		EList<EObject> elementList = correctLibrary.getReuseElementList();
		for(EObject libraryElement : elementList) {
			String name = getReuseElementName(libraryElement);
			if(!name.equals(currentNameInLibrary)) {
				currentExistingLibraryElementNames.add(name);
			}
		}

		// Ask for element name (when updating element in library).
		final InputDialog updateDialog =
				new InputDialog(Display.getCurrent().getActiveShell(), SHELL_TITLE,
						"Enter the name with which the element should be updated in the library '" +
								getReuseElementName(correctLibrary) +
								"':\n(Current name in library: " + currentNameInLibrary + ")\n ",
						currentNameInLibrary, new UpdatedNameValidator());
		if(updateDialog.open() != OK) {
			return;
		}
		String updateName = updateDialog.getValue();

		// Execute "updating" action (if the selected element still exists).
		if(doesReuseElementExistWithContext(selectedElement)) {
			ICommandStackService.getInstance().runAsCommand(selectedElement, () -> {
				updateElemInLib(selectedElement, correctLibrary, correctLibrarySpec, updateName);
			});
		} else {
			warnOfDeletedReuseElem(ACTION_DESCRIPTION, SHELL_TITLE);
			return;
		}

		raiseAndUpdateReuseLibraryView();
	}

	/**
	 * Updates the copy of the given reuse element in the specified
	 * {@link ReuseLibrary} with its new element version. The
	 * {@link ReuseElementSpec} is the one that connects reuse element and
	 * {@link ReuseLibrary}.
	 */
	private void updateElemInLib(EObject newElement, ReuseLibrary library,
			ReuseElementSpec referencingSpec, String newName) {
		if(library == null || newElement == null || referencingSpec == null || newName == null) {
			return;
		}

		EObject originalLibraryElement =
				checkForElementInLibrary(library, referencingSpec, SHELL_TITLE);
		if(originalLibraryElement == null) {
			return;
		}
		replaceElementInLibrary(originalLibraryElement, newElement, referencingSpec, newName);

		try {
			saveReuseLibrary(library);
		} catch(Exception e) {
			openError(getActiveWorkbenchWindow().getShell(), "Error in '" + SHELL_TITLE + "'",
					e.getMessage());
		}

		PrototypeService.getInstance().updatePrototypes();

		String currentActionType = "updateInLibrary";
		manageExternalReferencesOfReuseElement(newElement, currentActionType, SHELL_TITLE);
	}

	/** Validator for the name of the updated element. */
	private class UpdatedNameValidator implements IInputValidator {

		/** {@inheritDoc} */
		@Override
		public String isValid(String newText) {
			if(newText.equals("")) {
				return "A name needs to be specified.";
			}
			for(String name : currentExistingLibraryElementNames) {
				if(newText.equals(name)) {
					return "This name already exists. Please use another.";
				}
			}
			return null;
		}
	}
}
