/*-------------------------------------------------------------------------+
| Copyright 2023 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.variability.analysis;

import static org.eclipse.emf.ecore.util.EcoreUtil.copy;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.pickFirstInstanceOf;
import static org.fortiss.variability.model.VariabilityModelElementFactory.createAndPC;

import java.util.List;

import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.base.model.element.IModelElementSpecification;
import org.fortiss.tooling.ext.variability.model.OptionalVariationPointSpecification;
import org.fortiss.tooling.ext.variability.service.IDependingElementProviderService;
import org.fortiss.variability.analysis.EMFProductLineTranslation;
import org.fortiss.variability.model.presence.AndPC;
import org.fortiss.variability.model.presence.PresenceCondition;
import org.fortiss.variability.model.presence.PresenceConditionTerm;

/**
 * Product-line translation with adaptation for different presence condition location in the tooling
 * kernel.
 * 
 * @author bayha
 */
public class OptimizedProductLineTranslation extends EMFProductLineTranslation {

	/** {@inheritDoc} */
	@Override
	protected PresenceConditionTerm getLocalPresenceCondition(EObject model) {
		if(model instanceof IModelElement) {

			List<IModelElementSpecification> specifications =
					((IModelElement)model).getSpecifications();

			OptionalVariationPointSpecification varSpec =
					pickFirstInstanceOf(OptionalVariationPointSpecification.class, specifications);

			if(varSpec == null) {
				return null;
			}

			PresenceCondition presenceCondition = varSpec.getPresenceCondition();

			if(presenceCondition == null) {
				return null;
			}

			return presenceCondition.resolveToFeatureLiterals();
		}

		return null;
	}

	/** {@inheritDoc} */
	@Override
	protected void collectEObject(EObject eo, EClass cls, PresenceConditionTerm pc) {
		class2TranslatedEObjects.add(cls, eo);

		collectPresenceCondition(eo, pc);

		// Also collect depending elements for the tooling kernel implementation.
		List<EObject> dependingElements =
				IDependingElementProviderService.getInstance().getDependingElements(eo);
		for(EObject depObj : dependingElements) {
			PresenceConditionTerm depObjPC = getLocalPresenceCondition(depObj);

			if(depObjPC != null) {
				depObjPC = createAndPC(copy(depObjPC), copy(pc), "");
			} else {
				depObjPC = pc;
			}

			collectEObject(depObj, getTranslatedClass(depObj.eClass()), depObjPC);
		}
	}

	/**
	 * Accumulates the given presence condition by conjugating it. This is necessary in case an
	 * element "inherits" more than one presence conditions due to dependencies from other elements.
	 */
	private PresenceConditionTerm collectPresenceCondition(EObject eObject,
			PresenceConditionTerm pc) {
		PresenceConditionTerm oldPC = object2presenceCondition.get(eObject);

		if(oldPC != null) {
			if(pc != null) {
				AndPC newPC = createAndPC(copy(oldPC), copy(pc), "");

				object2presenceCondition.put(eObject, newPC);
				return newPC;
			}

			object2presenceCondition.put(eObject, oldPC);
			return oldPC;
		}

		object2presenceCondition.put(eObject, pc);
		return pc;
	}
}
