/*-------------------------------------------------------------------------+
| Copyright 2023 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.extension.base;

import static org.conqat.lib.commons.collections.CollectionUtils.asUnmodifiable;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.OptionalInt;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.kernel.extension.IPrototypeProvider;
import org.fortiss.tooling.kernel.extension.data.Prototype;
import org.fortiss.tooling.kernel.extension.data.PrototypeCategory;

/**
 * Base implementation for {@link IPrototypeProvider}s.
 * This implementation is intended for <i>static</i> prototype lists, i.e., lists which are never
 * changing in the course of the execution. If you want dynamic lists of prototypes, implement
 * directly {@link IPrototypeProvider}.
 * <p>
 * Sub-classes must implement {@link #registerPrototypes()} to register specific prototypes using
 * {@link #registerPrototype(String, EObject, String)} and
 * {@link #registerPrimaryPrototype(String, EObject, String)}.
 * 
 * @author hoelzl
 */
public abstract class PrototypeProviderBase implements IPrototypeProvider {

	/** Default value for expansion of prototype categories. */
	private static final boolean CATEGORY_EXPANDED_BY_DEFAULT = false;

	/** Stores the registered prototypes. */
	protected final List<Prototype> prototypes = new ArrayList<Prototype>();

	/** Stores the prototypes categories. */
	protected final Map<String, PrototypeCategory> categories =
			new HashMap<String, PrototypeCategory>();

	/** Constructor. */
	public PrototypeProviderBase() {
		registerPrototypes();
	}

	/** {@inheritDoc} */
	@Override
	public final void refreshCurrentPrototypes() {
		prototypes.clear();
		registerPrototypes();
	}

	/** Registers all prototypes. Sub-classes must override. */
	protected abstract void registerPrototypes();

	/** Registers the given {@link EObject} with the given name and category. */
	protected final void registerPrototype(String name, EObject prototype, String categoryName) {
		registerPrototype(name, prototype, categoryName, false, OptionalInt.empty());
	}

	/** Registers the given {@link EObject} with the given name, category, and priority. */
	protected final void registerPrototype(String name, EObject prototype, String categoryName,
			int priority) {
		registerPrototype(name, prototype, categoryName, false, OptionalInt.of(priority));
	}

	/**
	 * Registers the given {@link EObject} with the given name as primary
	 * prototype with the given category.
	 */
	protected final void registerPrimaryPrototype(String name, EObject prototype,
			String categoryName) {
		registerPrototype(name, prototype, categoryName, true, OptionalInt.empty());
	}

	/**
	 * Registers the given {@link EObject} with the given name and priority as primary prototype
	 * with the given category.
	 */
	protected final void registerPrimaryPrototype(String name, EObject prototype,
			String categoryName, int priority) {
		registerPrototype(name, prototype, categoryName, true, OptionalInt.of(priority));
	}

	/** Base method that actually performs the registration of prototypes. */
	private final void registerPrototype(String name, EObject prototype, String categoryName,
			boolean primary, OptionalInt priority) {
		Prototype prototypeObject;
		if(priority.isPresent()) {
			prototypeObject = new Prototype(name, prototype, primary, priority.getAsInt());
		} else {
			prototypeObject = new Prototype(name, prototype, primary);
		}
		prototypes.add(prototypeObject);
		registerElementInCategory(prototypeObject, categoryName);
	}

	/** {@inheritDoc} */
	@Override
	public final List<Prototype> getPrototypes() {
		return asUnmodifiable(prototypes);
	}

	/** Registers the element with the given category. */
	private void registerElementInCategory(Prototype element, String category) {
		PrototypeCategory cat = registerPrototypeCategory(category, CATEGORY_EXPANDED_BY_DEFAULT);
		// prevent duplicates within a category:
		// in case of already existing prototype, overwrite it
		for(Object object : cat.getChildren()) {
			if(object instanceof Prototype) {
				Prototype prototype = (Prototype)object;
				if(prototype.getName().equals(element.getName())) {
					cat.replace(prototype, element);
					return;
				}
			}
		}
		cat.add(element);
	}

	/** Registers an existing prototype category. */
	protected PrototypeCategory registerPrototypeCategory(String category,
			boolean expandByDefault) {
		PrototypeCategory cat = categories.get(category);
		if(cat == null) {
			cat = new PrototypeCategory(category);
			cat.setExpandedByDefaultStatus(expandByDefault);
			categories.put(category, cat);
		}
		return cat;
	}

	/** Registers an existing prototype category as child of a parent category. */
	protected PrototypeCategory registerPrototypeCategory(PrototypeCategory parent, String category,
			boolean expandByDefault) {
		PrototypeCategory cat = registerPrototypeCategory(category, expandByDefault);
		parent.add(cat);
		cat.setPrototypeCategoryParent(parent);
		return cat;
	}

	/**
	 * Sets the priority of the given prototype category. If the category does not exist, a
	 * {@link IllegalArgumentException} is thrown. Only needed if the category should be sorted
	 * based on the new priority and not by name.
	 */
	protected void setPrototypeCategoryPriority(String category, int priority)
			throws IllegalArgumentException {
		PrototypeCategory cat = categories.get(category);
		if(cat == null) {
			throw new IllegalArgumentException(
					"Cannot set category priority since no prototype category exists with the given name '" +
							category + "'.");
		}
		cat.setPriority(priority);
	}

	/** {@inheritDoc} */
	@Override
	public List<PrototypeCategory> getCategories() {
		List<PrototypeCategory> result = new ArrayList<PrototypeCategory>();
		result.addAll(categories.values());
		return result;
	}

	/** {@inheritDoc} */
	@Override
	public final String getUniqueID(Prototype prototype) {
		if(!prototypes.contains(prototype)) {
			return null;
		}
		String categoryName = "";
		for(PrototypeCategory category : this.getCategories()) {
			for(Object object : category.getChildren()) {
				if(object instanceof Prototype && ((Prototype)object).equals(prototype)) {
					categoryName = category.getName();
					break;
				}
			}
			if(!categoryName.isBlank()) {
				break;
			}
		}
		return this.getClass().getName() + '/' + categoryName + '/' + prototype.getName();
	}

	/** {@inheritDoc} */
	@Override
	public final Prototype getPrototypeByID(String id) {
		String[] idSplit = id.split("/");
		// should contain three parts:
		// 1. prototypeProvider name
		// 2. category name (in some cases empty)
		// 3. prototype name (either connected to a category or not)
		if(idSplit.length != 3) {
			return null;
		}
		String className = this.getClass().getName();
		if(!idSplit[0].equals(className)) {
			return null;
		}
		if(!idSplit[1].isBlank() && categories.containsKey(idSplit[1])) {
			PrototypeCategory category = categories.get(idSplit[1]);
			for(Object object : category.getChildren()) {
				if(object instanceof Prototype) {
					Prototype prototype = (Prototype)object;
					if(prototype.getName().equals(idSplit[2])) {
						return prototype;
					}
				}
			}
		} else {
			// in some cases (e.g. for mode switch specifications), the overall category could not
			// be found, but their prototypes are still in the list of all prototypes (and can be
			// found just with their names)
			for(Prototype prototype : prototypes) {
				if(prototype.getName().equals(idSplit[2])) {
					return prototype;
				}
			}
		}
		return null;
	}
}
