/*-------------------------------------------------------------------------+
| Copyright 2023 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.extension.data;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

/**
 * Container object for prototypes of a certain category.
 * 
 * @author doebber
 */
public class PrototypeCategory {

	/** Stores the name of the category. */
	private final String categoryName;

	/** Stores the parent of the category. */
	private PrototypeCategory parentCategory;

	/** Stores the priority of the category. Is important for sorting: lower value = lower prio. */
	private int categoryPriority;

	/**
	 * Stores the status whether the category should be expanded by default in Model Elements
	 * Navigator.
	 */
	private boolean expandedByDefault;

	/** Stores the prototypes affiliated with this category. */
	private List<Prototype> children = new ArrayList<Prototype>();

	/** Stores the unique sub-categories of this category. */
	private Set<PrototypeCategory> subCategories = new HashSet<PrototypeCategory>();

	/** Constructor. */
	public PrototypeCategory(String category) {
		categoryName = category;
		categoryPriority = 0;
		expandedByDefault = false;
	}

	/** Returns the priority of the category. */
	public int getPriority() {
		return categoryPriority;
	}

	/** Sets the priority of the category. */
	public void setPriority(int priority) {
		categoryPriority = priority;
	}

	/**
	 * Returns whether the category should be expanded (true) or not (false) by default in
	 * Model Elements Navigator.
	 */
	public boolean getExpandedByDefaultStatus() {
		return expandedByDefault;
	}

	/**
	 * Sets the category to be expanded (true) or not (false) by default in Model Elements
	 * Navigator.
	 */
	public void setExpandedByDefaultStatus(boolean shouldBeExpanded) {
		expandedByDefault = shouldBeExpanded;
	}

	/** Adds the prototype. */
	public void add(Prototype prototype) {
		children.add(prototype);
	}

	/** Adds sub-category. */
	public void add(PrototypeCategory subCat) {
		subCategories.add(subCat);
	}

	/** Replaces an existing prototype with a new one. */
	public void replace(Prototype oldPrototype, Prototype newPrototype) {
		int index = children.indexOf(oldPrototype);
		if(index != -1) {
			children.set(index, newPrototype);
		}
	}

	/** Returns the members of this category. */
	public Object[] getChildren() {
		List<Object> result = new ArrayList<Object>();
		result.addAll(children);
		result.addAll(subCategories);
		return result.toArray();
	}

	/** Returns the name of the category. */
	public String getName() {
		return categoryName;
	}

	/** Sets the parent of this category. */
	public void setPrototypeCategoryParent(PrototypeCategory parent) {
		parentCategory = parent;
	}

	/** Returns parent of this category. */
	public PrototypeCategory getParentCategory() {
		return parentCategory;
	}

	/** {@inheritDoc} */
	@Override
	public String toString() {
		return "Prototype category: " + categoryName;
	}

	/** {@inheritDoc} */
	@Override
	public int hashCode() {
		return getName().hashCode();
	}
}
