/*-------------------------------------------------------------------------+
| Copyright 2023 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.variability.analysis;

import java.util.HashMap;
import java.util.Map;

import org.eclipse.emf.ecore.EClass;

import com.microsoft.z3.BoolExpr;
import com.microsoft.z3.EnumSort;
import com.microsoft.z3.Expr;

/**
 * Base class for product line constraints to be checked by the product-line analysis.
 * 
 * All constraints are quantified expressions of the form:
 * [ForAll|Exists] <list of quantifed variables>: <body using quantified variables>
 * 
 * @author bayha
 */
public abstract class ProductLineConstraintBase implements IProductLineConstraint {
	/** Prefix to label tracking variables. */
	private static final String TRACK_PREFIX = "TRACK_";

	/** The body of this constraint. I.e. the quantified expression. */
	protected BoolExpr body;

	/** The {@link EClass} types of the quantified variables. */
	protected EClass[] variableTypes;

	/** The names of the quantified variables. */
	protected String[] variableNames;

	/** The quantified variables */
	protected Expr<?>[] variables;

	/**
	 * The tracking constants which are used to extract violating interpretations for the quantifier
	 * variables.
	 */
	protected Expr<?>[] trackerVars;

	/** A map from variable names to expressions which can be used in the body. */
	protected Map<String, Expr<?>> name2Variables = new HashMap<String, Expr<?>>();

	/** The {@link IProductLineTranslation} to be used for creation of variables. */
	protected IProductLineTranslation translation;

	/** Constructor. */
	public ProductLineConstraintBase(IProductLineTranslation translation) {
		this.translation = translation;

		variableTypes = createVariablesClasses();
		variableNames = createVariableNames();
		trackerVars = new Expr[variableNames.length];
	}

	/** Creates the actual body for this constraint using the quantified variables. */
	protected abstract BoolExpr createBody();

	/** Specifies the names for the quantified variables. */
	protected abstract String[] createVariableNames();

	/** Specifies the EClass types for the quantified variables. */
	protected abstract EClass[] createVariablesClasses();

	/** {@inheritDoc} */
	@Override
	public final Expr<?>[] getQuantifierVariables() {
		if(variables == null) {
			createVariables();
		}
		return variables;
	}

	/** {@inheritDoc} */
	@Override
	public final Expr<?>[] getVariableTracker() {
		return trackerVars;
	}

	/** {@inheritDoc} */
	@Override
	public final BoolExpr getBody() {
		if(variables == null) {
			createVariables();
		}

		if(body == null) {
			body = createBody();
			createTracking();
		}

		return body;
	}

	/** {@inheritDoc} */
	@Override
	public final EClass[] getVariablesClasses() {
		return variableTypes;
	}

	/** Creates tracking variables and adds them to the body. */
	private final void createTracking() {
		int numVariables = variableNames.length;
		BoolExpr[] trackExprs = new BoolExpr[numVariables + 1];

		for(int i = 0; i < variables.length; i++) {
			trackExprs[i] = translation.getContext().mkEq(trackerVars[i], variables[i]);
		}

		trackExprs[numVariables] = body;

		body = translation.getContext().mkAnd(trackExprs);
	}

	/** Creates the Z3 variables for this constraint, using the translation. */
	private final void createVariables() {
		int numVariables = variableTypes.length;
		variables = new Expr[numVariables];

		for(int i = 0; i < numVariables; i++) {
			EClass ec = variableTypes[i];
			EnumSort<?> ecSort = translation.getDatatype(ec);

			String varName = variableNames[i];
			trackerVars[i] = translation.getContext().mkConst(TRACK_PREFIX + varName, ecSort);

			Expr<?> var = translation.getContext().mkConst(varName, ecSort);
			variables[i] = var;
			name2Variables.put(varName, var);
		}
	}
}
