/*-------------------------------------------------------------------------+
| Copyright 2021 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.variability.bind;

import static java.util.Arrays.asList;
import static java.util.Collections.singletonList;
import static org.fortiss.variability.model.VariabilityModelElementFactory.getRemoveBindingStrategyInstance;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.fortiss.variability.model.IBindingStrategy;
import org.fortiss.variability.model.IVariationPoint;
import org.fortiss.variability.model.base.RemoveBindingStrategy;
import org.fortiss.variability.model.features.AbstractAlternativeFeature;
import org.fortiss.variability.model.features.AbstractCompositionalFeature;
import org.fortiss.variability.model.features.AbstractCrossFeatureConstraint;
import org.fortiss.variability.model.features.AbstractFeature;
import org.fortiss.variability.model.features.configuration.AlternativeFeatureConfiguration;
import org.fortiss.variability.model.features.configuration.CompositionalFeatureConfiguration;
import org.fortiss.variability.model.features.configuration.IFeatureConfiguration;
import org.fortiss.variability.model.features.configuration.VariantConfiguration;
import org.fortiss.variability.util.Pair;

/**
 * Utility functions for implementing variation point binding.
 * 
 * @author bayha
 */
public class BindUtils {
	/**
	 * Configures the given {@link EObject} for the given
	 * {@link VariantConfiguration} according to the given {@link IBindingStrategy}.
	 * 
	 * Hereby also all contained elements are recursively configured.
	 * 
	 * @param eObj
	 *            The {@link EObject} to be configured.
	 * @param varConf
	 *            The {@link VariantConfiguration} to be used for
	 *            configuration.
	 * @param strategy
	 *            The {@link IBindingStrategy} to be used for executing the
	 *            binding of elements itself.
	 * 
	 * @return The configured {@link EObject} with configured children.
	 */
	public static <T extends EObject> T configureEObject(T eObj, VariantConfiguration varConf,
			IBindingStrategy strategy) {
		if(eObj instanceof IVariationPoint<?>) {
			EObject evaluatedObj = (EObject)((IVariationPoint<?>)eObj).bind(varConf, strategy);

			if(evaluatedObj == null) {
				return null;
			}
		}

		EList<EObject> contents = eObj.eContents();
		for(int i = contents.size() - 1; i >= 0; i--) {
			if(i >= contents.size()) {
				continue;
			}
			configureEObject(contents.get(i), varConf, strategy);
		}

		return eObj;
	}

	/**
	 * Configures the given {@link EObject} for the given
	 * {@link VariantConfiguration}. Here the {@link RemoveBindingStrategy} is used.
	 * 
	 * Hereby also all contained elements are recursively configured.
	 * 
	 * @param eObj
	 *            The {@link EObject} to be configured.
	 * @param varConf
	 *            The {@link VariantConfiguration} to be used for configuration.
	 * 
	 * @return The configured {@link EObject} with configured children.
	 */
	protected static <T extends EObject> T configureEObject(T eObj, VariantConfiguration varConf) {
		return configureEObject(eObj, varConf, getRemoveBindingStrategyInstance());
	}

	/**
	 * Recursive implementation of the completeness and correctness check for
	 * variant configurations.
	 */
	private static Pair<String, List<EObject>> checkVariantConfiguration(
			IFeatureConfiguration<?> conf, AbstractFeature feature,
			Set<AbstractFeature> selectedFeatures,
			Map<AbstractFeature, AbstractFeature> requiredFeatures,
			Map<AbstractFeature, AbstractFeature> excludedFeatures) {

		for(AbstractCrossFeatureConstraint cnstr : feature.getConstraints()) {
			final AbstractFeature cnstrTarget = cnstr.getTarget();

			switch(cnstr.getType()) {
				case EXCLUDES:
					excludedFeatures.put(feature, cnstrTarget);
					break;
				case REQUIRES:
					requiredFeatures.put(feature, cnstrTarget);
					break;
				default:
					System.err.println("Unknown Constraint \"" + cnstr.getClass() + "\" found.");
					break;
			}
		}

		if(feature instanceof AbstractCompositionalFeature) {
			AbstractCompositionalFeature compFeature = (AbstractCompositionalFeature)feature;

			if(!(conf instanceof CompositionalFeatureConfiguration)) {
				System.err.println("A feature configuration has an invalid configuration of type " +
						conf.getClass().getCanonicalName() + " for the compositional feature " +
						feature.getName());
				return null;
			}

			final CompositionalFeatureConfiguration compConf =
					(CompositionalFeatureConfiguration)conf;

			Pair<String, List<EObject>> res = checkVariantConfigurationForCompositionalFeature(
					selectedFeatures, requiredFeatures, excludedFeatures, compFeature, compConf);
			if(res != null) {
				return res;
			}
		}
		if(feature instanceof AbstractAlternativeFeature) {
			final AbstractAlternativeFeature altFeature = (AbstractAlternativeFeature)feature;

			if(!(conf instanceof AlternativeFeatureConfiguration)) {
				System.err.println("A feature configuration has an invalid configuration of type " +
						conf.getClass().getCanonicalName() + " for the alternative feature " +
						feature.getName());
				return null;
			}

			final AlternativeFeatureConfiguration altConf = (AlternativeFeatureConfiguration)conf;
			final IFeatureConfiguration<?> altSelectedConf = altConf.getSelectedConfiguration();

			Pair<String, List<EObject>> res =
					checkVariantConfigurationForAlternativeFeature(selectedFeatures,
							requiredFeatures, excludedFeatures, altFeature, altSelectedConf);
			if(res != null) {
				return res;
			}
		}

		// Check constraints
		for(AbstractFeature source : requiredFeatures.keySet()) {
			final AbstractFeature reqFeat = requiredFeatures.get(source);
			if(selectedFeatures.contains(source) && !selectedFeatures.contains(reqFeat)) {
				String msg = "The selected feature \"" + source.getName() +
						"\" requires the feature \"" + reqFeat.getName() +
						"\" to be selected, too.";
				return new Pair<String, List<EObject>>(msg, asList(reqFeat, source));
			}
		}
		for(AbstractFeature source : excludedFeatures.keySet()) {
			final AbstractFeature exclFeat = excludedFeatures.get(source);
			if(selectedFeatures.contains(source) && selectedFeatures.contains(exclFeat)) {
				String msg = "The selected feature \"" + source.getName() +
						"\" excludes the feature \"" + exclFeat.getName() +
						"\" to be selected at the same time.";
				return new Pair<String, List<EObject>>(msg, asList(exclFeat, source));
			}
		}
		return null;
	}

	/**
	 * Part of the recursive implementation of the completeness and correctness check for variant
	 * configurations for alternative features.
	 */
	private static Pair<String, List<EObject>> checkVariantConfigurationForAlternativeFeature(
			Set<AbstractFeature> selectedFeatures,
			Map<AbstractFeature, AbstractFeature> requiredFeatures,
			Map<AbstractFeature, AbstractFeature> excludedFeatures,
			final AbstractAlternativeFeature altFeature,
			final IFeatureConfiguration<?> altSelectedConf) {
		if(!altFeature.isOptional() && altSelectedConf == null) {
			final String msg = "The alternative feature \"" + altFeature.getName() +
					"\" is not optional, but has no alternative selected.";
			return new Pair<String, List<EObject>>(msg, singletonList(altFeature));
		}

		if(altSelectedConf != null) {
			final AbstractFeature altSelectedFeature = altSelectedConf.getFeatureReference();
			selectedFeatures.add(altSelectedFeature);

			final Pair<String, List<EObject>> altResult = checkVariantConfiguration(altSelectedConf,
					altSelectedFeature, selectedFeatures, requiredFeatures, excludedFeatures);
			if(altResult != null) {
				return altResult;
			}
		}

		return null;
	}

	/**
	 * Part of the recursive implementation of the completeness and correctness check for variant
	 * configurations for compositional features.
	 */
	private static Pair<String, List<EObject>> checkVariantConfigurationForCompositionalFeature(
			Set<AbstractFeature> selectedFeatures,
			Map<AbstractFeature, AbstractFeature> requiredFeatures,
			Map<AbstractFeature, AbstractFeature> excludedFeatures,
			AbstractCompositionalFeature compFeature,
			final CompositionalFeatureConfiguration compConf) {
		Map<AbstractFeature, IFeatureConfiguration<?>> feature2conf =
				new HashMap<AbstractFeature, IFeatureConfiguration<?>>();
		compConf.getSubConfigurations().stream()
				.forEach(c -> feature2conf.put(c.getFeatureReference(), c));

		for(AbstractFeature f : compFeature.getSubFeatures()) {
			final IFeatureConfiguration<?> fConf = feature2conf.get(f);

			if(!f.isOptional() && (fConf == null || !fConf.isSelected())) {
				String msg = "The feature \"" + f.getName() + "\" is mandatory but not selected.";
				if(f instanceof AbstractAlternativeFeature) {
					msg = "The feature \"" + f.getName() +
							"\" is mandatory but no alternative has been selected.";
				}
				return new Pair<String, List<EObject>>(msg, singletonList(f));
			}

			if(fConf != null && fConf.isSelected()) {
				selectedFeatures.add(f);

				final Pair<String, List<EObject>> fResult = checkVariantConfiguration(fConf, f,
						selectedFeatures, requiredFeatures, excludedFeatures);
				if(fResult != null) {
					return fResult;
				}
			}
		}

		return null;
	}

	/**
	 * Checks the given {@link IFeatureConfiguration} for correctness and
	 * completeness wrt the constraints in the feature model.
	 * 
	 * @param varConf
	 *            The {@link IFeatureConfiguration} to be checked.
	 * 
	 * @return {@link List} of Problem descriptions and problematic elements if
	 *         there should be any problems. 'null', else.
	 */
	public static Pair<String, List<EObject>>
			checkVariantConfiguration(VariantConfiguration varConf) {
		HashSet<AbstractFeature> selected = new HashSet<AbstractFeature>();
		HashMap<AbstractFeature, AbstractFeature> required =
				new HashMap<AbstractFeature, AbstractFeature>();
		HashMap<AbstractFeature, AbstractFeature> excluded =
				new HashMap<AbstractFeature, AbstractFeature>();

		final Pair<String, List<EObject>> result = checkVariantConfiguration(varConf,
				varConf.getFeatureModel(), selected, required, excluded);
		return result;
	}
}
